from nlpe import EvaluatorProxy
import numpy as np
    
class Evaluator:
    def dump(self):
        print("dump evaluator")
        
    def compute(self, predictions, label):
        assert len(predictions) == len(label)
        length = len(predictions)
        acc = np.sum(np.array(predictions) == np.array(label))/length
        return {
            "acc": acc,
        }
        
            
def test_evaluator_proxy():
    evaluator = Evaluator()
    proxy = EvaluatorProxy(
        glossary="evaluator_ proxy",
        compute_call= evaluator.compute
    )
    
    assert 0.33 == round(proxy.compute(predictions=[1,2,3], label=[1,1,1])["acc"],2)


from abc import ABC, abstractmethod
from enum import StrEnum, auto
from typing import Any, Callable, List, Optional, Union
from ..utils import global_logger
from .data import Data


class Language(StrEnum):
    ENGLISH = auto()
    JAPANESE = auto()
    CHINESE = auto()


class Text(str):        
    def __init__(self, *args, language: Language = Language.ENGLISH, **kwargs) -> None:
        super().__init__()
        assert isinstance(language, Language)
        self._language = language
    
    def tokens(self, tokenizor: Callable, **kwargs) -> List[Any]:
        assert isinstance(tokenizor, Callable)
        return tokenizor(self)
    
    def embeddings(self, encoder: Callable, **kwargs) -> Union[Any, List[Any]]:
        assert isinstance(encoder, Callable)
        return encoder(self)   
    
    @property
    def language(self):
        return self._language
    
    @language.setter
    def language(self, value):
        assert isinstance(value, Language) or Language(value).value == value
        self._language = value 
        
         
class Sentence(Text):
    pass
    

class TextData(Data):
    def __init__(self, map_dataset_to_text_list: Callable, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(map_dataset_to_text_list, Callable)
        self._map_dataset_to_text_list = map_dataset_to_text_list
        self._max_length = None
        self._min_length = None

    @property
    def all_texts(self) -> List[Text]:
        result = []
        for ds in self.load_dataset():
            text_list = self._map_dataset_to_text_list(ds)
            assert isinstance(text_list, List)
            for t in text_list:
                assert isinstance(t, Text)
                result.append(t)
        return result
    
    def statistic_all_texts(self, tokenizor: Callable) -> List[int]:
        all_lengthes = [len(t.tokens(tokenizor)) for t in self.all_texts]
        self._min_length = min(all_lengthes)
        self._max_length = max(all_lengthes)
        return all_lengthes
       
    @property
    def min_length(self) -> Optional[int]:
        if self._min_length is None:
            logger = global_logger()
            logger.warning("Should not use min_length before statisticing the all datasets!")
        return self._min_length
    
    @property
    def max_length(self) -> Optional[int]:
        if self._max_length is None:
            logger = global_logger()
            logger.warning("Should not use min_length before statisticing the all datasets!")
        return self._max_length
    


from abc import ABC, abstractmethod
from typing import Callable, Dict, Any, Union, Optional, List, Tuple, Iterable

from ..data import Data
from ..proxy import ProxyBase
from ..utils.utils import normalize_str_arg
from ..utils.glossary import Glossary
from ..argument import ArgumentPool


class Approach(ABC):
    def __init__(self, *args, glossary: Union[str, Glossary] = None, **kwargs):
        if glossary is None:
            glossary = ArgumentPool().meta_argument["approach"]
        if isinstance(glossary, str):
            glossary = Glossary(normalize_str_arg(glossary))
        assert isinstance(glossary, Glossary)
        self._glossary = glossary
        self._processing_data: Data = None


    @property
    def processing_data(self) -> Data:
        return self._processing_data
    
    @property
    def glossary(self) -> Glossary:
        return self._glossary

    @abstractmethod
    def _process(self, data: Data, *args, **kwargs):
        pass

    def process(self, data: Data, *args, **kwargs):
        assert isinstance(data, Data)
        self._processing_data = Data
        self._process(data, *args, **kwargs)
        self._processing_data = None




import logging
import os
from datetime import datetime
from pathlib import Path, PurePath

_global_logger = None
import socket

def global_logger(log_dir: str = None, logger_name: str = None) -> logging.Logger:
    global _global_logger
    if isinstance(_global_logger, logging.Logger):
        return _global_logger

    if not isinstance(log_dir, str) or len(log_dir) == 0:
        log_dir = os.environ.get("LOG_DIR", f"./log/{socket.gethostname()}")
    if not isinstance(logger_name, str) or len(logger_name) == 0:
        logger_name = os.environ.get("PROJECT_NAME", "nlpe")
    os.makedirs(log_dir, exist_ok=True)
    ph = PurePath(log_dir, f"{logger_name}_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log")
    formatter = logging.Formatter("[%(asctime)s - %(name)s - %(levelname)s] %(message)s")

    _global_logger = logging.getLogger(logger_name)
    _global_logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(str(ph))
    file_handler.setFormatter(formatter)
    _global_logger.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    _global_logger.addHandler(console_handler)

    return _global_logger



from copy import deepcopy
import sys


ORIGINAL_SYS_ARGV = deepcopy(sys.argv)


def extend_to_sys_argv(string):
    sys.argv.extend(string.split())
    

def extend_to_original_sys_argv(string):
    sys.argv = deepcopy(ORIGINAL_SYS_ARGV)
    extend_to_sys_argv(string)



from copy import deepcopy
import dataclasses
from enum import Enum, StrEnum, auto
from pathlib import Path
import pytest
import sys
import argparse

# from nltk import EarleyChartParser
from nlpe import ArgumentPool, ArgumentFactory
from nlpe.utils.test import extend_to_original_sys_argv, extend_to_sys_argv

# import pip
# installed_packages = pip.get_installed_distributions()
# installed_packages_list = sorted(["%s==%s" % (i.key, i.version)
#      for i in installed_packages])
# print(installed_packages_list)


def setup_function(function):
    extend_to_original_sys_argv("-t test_task -d test_dataset -a test_approach --debug")

def test_single_argument_pool():
    assert id(ArgumentPool()) ==  id(ArgumentPool())

    
def test_meta_argument():
    pool = ArgumentPool()
    pool.pop(pool.meta_argument_token)
    factory = pool.meta_argument_factory
    assert isinstance(pool.meta_argument, dict)
    assert isinstance(factory, ArgumentFactory)
    assert pool.meta_argument == factory.to_dict(factory.argument)
    assert pool.meta_argument["task"] == "test_task"
    assert pool.meta_argument["dataset"] == "test_dataset"
    assert pool.meta_argument["approach"] == "test_approach"
    current_size = pool.size
    pool.pop(pool.meta_argument_token)
    assert pool.size == current_size - 1
    pool.meta_argument
    assert pool.size == current_size
    # Path(pool.meta_argument["dataset_raw_dir"]).mkdir(parents=True, exist_ok=True)
    
    
def test_parse_argument():
    class test_enum(StrEnum):
        TEST = auto()
        
    parser = argparse.ArgumentParser()
    parser.add_argument("--int_arg", type=int, required=True)
    parser.add_argument("--float_arg", type=float, required=True)
    parser.add_argument("--enum_arg", type=test_enum, required=True)
    
    
    extend_to_sys_argv("--int_arg 1 --float_arg 1.5 --enum_arg test")
    
    factory = ArgumentFactory(process = lambda : parser.parse_known_args()[0], argument_glossary="argparser", argument_type=argparse.Namespace, to_dict=vars)
    
    pool = ArgumentPool()
    current_size = pool.size
    token = pool.push(factory)
    assert factory.argument_type == argparse.Namespace
    assert isinstance(factory.argument, factory.argument_type)
    assert id(factory) == id(pool.search(token))
    factory = pool.search(token)
    assert factory.argument_dict["int_arg"] == 1
    assert factory.argument_dict["float_arg"] == 1.5
    assert factory.argument_dict["enum_arg"] == test_enum.TEST
    assert pool.size == current_size + 1


from nlpe import Pool, POOL_DEFAULT_CAPACITY, UnifiedToken, TokenStatus

class MyUnit:
    pass

class MyPool(Pool):
    def __init__(self, unit_type: type = MyUnit):
        super().__init__(unit_type)
        
        

def test_basic():
    pool = MyPool()
    assert pool.size == 0
    assert POOL_DEFAULT_CAPACITY > 0
    assert isinstance(POOL_DEFAULT_CAPACITY, int)
    token2unit = dict()
    for i in range(POOL_DEFAULT_CAPACITY):
        assert pool.size == i
        unit = MyUnit()
        token2unit[pool.push(unit)] = unit
        assert pool.size == i + 1
    
    for t, u in token2unit.items():
        assert t.status == TokenStatus.VALID
        assert isinstance(pool.search(t), MyUnit)
        assert t.status == TokenStatus.VALID
        assert id(u) == id(pool.pop(t))
        assert t.status == TokenStatus.INVALID
    
    pool.push(unit)
    pool.push(unit)
    pool.push(unit)
    assert pool.size == 3
    assert pool.reset()
    assert pool.size == 0


from typing import Any, Callable

from ..utils.glossary import Glossary
from ..proxy import ProxyBase
from ..argument import ArgumentPool


class EvaluatorProxy(ProxyBase):
    def __init__(self, compute_call: Callable, glossary: Glossary=None) -> None:
        if glossary is None:
            glossary = ArgumentPool().meta_argument["dataset"] + "_evaluator"
        super().__init__(glossary)
        assert isinstance(compute_call, Callable)
        self._compute_call = compute_call
    
    def compute(self, *args, **kwargs) -> Any:
        return self._compute_call(*args, **kwargs)



import inspect
import os
from enum import StrEnum, auto
from pathlib import Path, PurePath
from typing import Optional, Tuple, Dict, Callable, Union, Any, List, Iterable

from ..argument import ArgumentPool, PATH_MODE
from ..utils import Glossary
from ..utils import singleton, global_logger, normalize_str_arg
from ..proxy import ProxyBase
from ..evaluator import EvaluatorProxy


class DatasetSplitCategory(StrEnum):
    TEST = 'test'
    TRAIN = 'train'
    VALIDATION = 'validation'

    @classmethod
    @property
    def all(cls):
        return tuple(DatasetSplitCategory)
    

class DatasetVersion(StrEnum):
    RAW = auto()
    VANILLA = auto()
    FEATURE = auto()


class TaskCategory(StrEnum):
    CLASSIFICATION = 'classification'
    GENERATION = 'generation'
    REGRESSIVE = "regressive"


class DatasetProxy(ProxyBase):
    def __init__(self, dataset_type: type, load_dataset_call: Callable, dump_dataset_call: Callable, raw_dir:Path = None, download_raw_call: Optional[Callable] = None, glossary: Union[str, Glossary] = None) -> None:
        if glossary is None:
            glossary = ArgumentPool().meta_argument["dataset"]
        super().__init__(glossary=glossary)
        
        assert isinstance(dataset_type, type)
        self._dataset_type = dataset_type
        assert isinstance(load_dataset_call, Callable)
        self._load_dataset_call = load_dataset_call
        if dump_dataset_call is None:
            dump_dataset_call = lambda *args, **kwargs: True
        assert isinstance(dump_dataset_call, Callable)
        self._dump_dataset_call = dump_dataset_call
        
        if isinstance(raw_dir, str):
            raw_dir = Path(normalize_str_arg(raw_dir))
        assert raw_dir is None or isinstance(raw_dir, Path)
        if not isinstance(raw_dir, Path):
            meta_argument = ArgumentPool().meta_argument
            raw_dir= Path(meta_argument["dataset_raw_dir"])
        self._raw_dir = raw_dir
        
        if download_raw_call is None:
            download_raw_call = lambda *args, **kwargs: True
        assert isinstance(download_raw_call, Callable)
        self._download_raw_call = download_raw_call
        self.download_raw()
        
    @property
    def dataset_type(self):
        return self._dataset_type

    def download_raw(self, force=False):
        raw_dir = self.raw_dir
        logger = global_logger()
        if raw_dir.is_dir():
            if force:
                logger.warning(f"Redownload dataset raw to dir '{raw_dir}'!")
                for root, dirs, files in self.data_dir().walk(top_down = False, on_error=print):
                    for f in files:
                        Path(f).unlink()
                    for d in dirs:
                        Path(d).rmdir()
                assert len(raw_dir.iterdir()) == 0
        else:
            raw_dir.mkdir(parents=True)
        assert raw_dir.is_dir()
        assert self._download_raw_call(raw_dir)

    def load_dataset(self, split: DatasetSplitCategory, *args, version: DatasetVersion = DatasetVersion.FEATURE, **kwargs) -> Any:
        assert isinstance(split, DatasetSplitCategory)
        result = self._load_dataset_call(self, split, *args, version=version, **kwargs)
        assert isinstance(result, self._dataset_type)
        return result
    
    def dump_dataset(self, split: DatasetSplitCategory, *args, version: DatasetVersion = DatasetVersion.FEATURE, **kwargs) -> bool:
        assert isinstance(split, DatasetSplitCategory)
        return self._dump_dataset_call(self, split, *args, version=version, **kwargs)
    
    @property
    def raw_dir(self) -> Path: 
        return self._raw_dir
    
class Data:
    def __init__(self, dataset_proxy:DatasetProxy, evaluator_proxy:EvaluatorProxy = None):
        assert isinstance(dataset_proxy, DatasetProxy)
        self._dataset_proxy = dataset_proxy
        self._dataset: Dict[DatasetSplitCategory, dataset_proxy.dataset_type] = None
        if evaluator_proxy is not None:
            assert isinstance(evaluator_proxy, EvaluatorProxy)
        self._evaluator_proxy = evaluator_proxy

    @property
    def dataset_proxy(self) -> DatasetProxy:
        return self._dataset_proxy
    
    @property
    def evaluator_proxy(self) -> EvaluatorProxy:
        return self._evaluator_proxy

    @property
    def dataset_name(self) -> str:
        return self._dataset_proxy.glossary.name
    
    @property
    def dataset_type(self):
        return self._dataset_proxy.dataset_type
    
    @property
    def all_datasets(self):
        return self.load_dataset()
    
    def __getitem__(self, key: DatasetSplitCategory):
        assert isinstance(key, DatasetSplitCategory)
        return self.load_dataset(key)
    
    def load_dataset(self, split: Union[DatasetSplitCategory, Tuple[DatasetSplitCategory]] = DatasetSplitCategory.all,  *args, **kwargs) -> Union[Any, Tuple[DatasetSplitCategory, Any]]:
        if isinstance(split, Tuple):
            result = []
            for s in split:
                assert isinstance(s, DatasetSplitCategory)
                result.append(self.load_dataset(s))
            result =  type(split)(result)

        elif isinstance(split, DatasetSplitCategory):
            if not isinstance(self._dataset, dict):
                self._dataset = dict()
            if split not in self._dataset:
                d = self._dataset_proxy.load_dataset(split)
                assert isinstance(d, self.dataset_type)
                self._dataset[split] = d
            result = self._dataset[split]
        elif split is None:
            result = None
        else:
            raise ValueError(f"Invaliad 'split': {split}")
        return result
    
    def dump_dataset(self, split: Union[DatasetSplitCategory, Tuple[DatasetSplitCategory]] = DatasetSplitCategory.all, *args, **kwargs) -> bool:
        result = True
        if isinstance(split, Tuple):
            for s in split:
                result = result and self.dump_dataset(s, *args, **kwargs)
        elif isinstance(split, DatasetSplitCategory):
            result = self._dataset_proxy.dump_dataset(split, *args, **kwargs)
        elif split is None:
            result = True
        else:
            raise ValueError(f"Invaliad 'split': {split}")
        return result
    
    
    def evaluate(self, prediction, label, *args, **kwargs):
        assert isinstance(self._evaluator_proxy, EvaluatorProxy)
        return self._evaluator_proxy.compute(prediction, label, *args, **kwargs)


from typing import Any, Callable, Union

from .utils.glossary import Glossary


class ProxyBase:
    def __init__(self, glossary: Union[str, Glossary]) -> None:
        if isinstance(glossary, str):
            glossary = Glossary(glossary)
            if not glossary.name.endswith("_proxy"):
                glossary = Glossary(glossary.name + "_proxy")
        assert isinstance(glossary, Glossary)
        self._glossary = glossary

        
    @property
    def glossary(self) -> Glossary:
        return self._glossary
    



# Singleton Pattern
def singleton(_class: type):
    class_names = {}
    name = repr(_class)

    def wrapped_class(*args, **kwargs):
        if name in class_names.keys():
            instance = class_names[name]
        else:
            instance = _class(*args, **kwargs)
            class_names[name] = instance
        return instance
    return wrapped_class


def init_argument_decorator(_decorator):
    def _decorator_wrap(cls):
        def _init_decorator(instance, *args, **kwargs):
            args, kwargs = _decorator(*args, **kwargs)
            return cls._origin__init__(instance, *args, **kwargs)
        cls._origin__init__ = cls.__init__
        cls.__init__ = _init_decorator
        return cls
    return _decorator_wrap





import argparse
import copy
import dataclasses
from functools import WRAPPER_ASSIGNMENTS, partial, update_wrapper, wraps
import json
import os
from pathlib import Path, PurePath
import sys
import uuid
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, fields, make_dataclass, Field, field
from enum import Enum, StrEnum, auto
from typing import Dict, Any, Set, Tuple, Union, Callable, Type, List, Optional, Iterable
from argparse import Namespace, ArgumentParser
from .utils import Glossary, GLOSSARY_POOL

from .utils import global_logger
import numpy as np

from .pool import Pool, UnifiedToken
from .utils.design_patterns import singleton



class PATH_MODE(StrEnum):
    ABSTRACT = 'abs'
    RELATED = 'rel'

    def __str__(self):
        return self.value


def _parsing_meta_arg() -> Namespace:
    arg_parser = ArgumentParser(usage="Initialize Meta Arguments", add_help=False)
    arg_parser.add_argument("-t", "--task", help="Specify the name of task.", type=str, required=True)
    arg_parser.add_argument("-d", "--dataset", help="Specify the dataset need to be deal with.", type=str, required=True)
    arg_parser.add_argument("-a", "--approach", help="Specify the approach", type=str, required=True)
    arg_parser.add_argument("--path_mode", help="Specify the mode of path. 'abs': the abstract path; 'rel': the related path to the root dir", type=PATH_MODE, default=PATH_MODE.ABSTRACT, choices=[pm.value for pm in PATH_MODE])
    arg_parser.add_argument("--cache_dir", help="Specify the dir of cached files, the default is '.cache' refer to root dir", type=PurePath, default=PurePath(".cache"))
    arg_parser.add_argument("--dataset_raw_dir", help="Specify the dir of raw version of dataset", type=PurePath, default=None)
    arg_parser.add_argument("--force_cache", help="Whether or not force to download cached info, such as dataset, configuration, when cached file is exist.", action='store_true')
    arg_parser.add_argument("--debug", help="Whether or not execuate by debug mode", action='store_true')
    argument = arg_parser.parse_known_args()[0]
    if argument.dataset_raw_dir is None:
        argument.dataset_raw_dir = PurePath(argument.cache_dir, "dataset", argument.dataset, "raw")
    return argument

class ArgumentFactory:
    def __init__(self,
            process: Callable, 
            argument_glossary: Glossary, 
            argument_type: Type,
            to_dict: Callable = None,
            process_args: Optional[Union[List,Tuple]] = None, 
            process_kw_args: Optional[Dict[str, Any]] = None,
            ) -> None:
        if isinstance(argument_glossary, str):
            argument_glossary = Glossary(argument_glossary)
        assert isinstance(argument_glossary, Glossary)
        self._argument_glossary = argument_glossary
        assert isinstance(process, Callable)
        assert isinstance(argument_type, Type)
        self._argument_type = argument_type
        if process_args is not None:
            assert isinstance(process_args, List) or isinstance(process_args, Tuple)
        else:
            process_args = list()
        if process_kw_args is not None:
            assert isinstance(process_kw_args, Dict) or isinstance(process_kw_args, Dict)
        else:
            process_kw_args = dict()
        self._process = update_wrapper(partial(process, *process_args, **process_kw_args), process)
        if issubclass(argument_type, dict) and to_dict is None:
            to_dict = lambda a, *args, **kwargs: dict(deepcopy(a))
        assert isinstance(to_dict, Callable)
        self._to_dict = to_dict
        self._argument = None
        
    @property
    def argument_glossary(self) -> Callable:
        return self._argument_glossary
    
    @property
    def process(self) -> Callable:
        return self._process
    
    @property
    def argument_type(self):
        return self._argument_type
    
    @property
    def to_dict(self):
        return self._to_dict
    
    @property
    def argument_dict(self) -> Any:
        result = self.to_dict(self.argument)
        assert isinstance(result, Dict)
        return result
    
    @property
    def argument(self) -> Any:
        if self._argument is None:
            self._argument = self.process()
        assert isinstance(self._argument, self.argument_type)
        return self._argument
    

@singleton
class ArgumentPool(Pool):
    def __init__(self):
        super().__init__(unit_type=ArgumentFactory)
        self._meta_argument_token: UnifiedToken = None
        logger = global_logger()
        meta_argument = self.meta_argument
        if meta_argument["debug"]:
            logger.debug("*** Meta Arguments ***")
            for n, v in meta_argument.items():
                logger.debug(f"{n}: {v}")
    
    @property
    def all_args(self) -> Dict[Glossary, Dict[str, Any]]:
        result = dict()
        for u in self._all_units:
            result[u.argument_glossary]= u.argument_dict
        return result

    def __getitem__(self, item:Glossary) -> Optional[Any]:
        if isinstance(item, str):
            item = GLOSSARY_POOL.search(item)
        assert isinstance(item, Glossary)
        factory = self.search(UnifiedToken(factor=item))
        if factory:
            return factory.argument
        raise KeyError
    
    @property
    def meta_argument(self) -> Dict[str, Any]:
        factory = self.meta_argument_factory
        result = factory.argument_dict
        return result
    
    @property
    def meta_argument_factory(self) -> ArgumentFactory:
        factory = self.search(self.meta_argument_token)
        assert isinstance(factory, ArgumentFactory)
        return factory

    @property
    def meta_argument_token(self) -> UnifiedToken:
        if not isinstance(self._meta_argument_token, UnifiedToken):
            argument_glossary = Glossary("meta_argument", force=True)
            factory = ArgumentFactory(process=_parsing_meta_arg, argument_glossary=argument_glossary, argument_type=Namespace, to_dict=vars)
            self._meta_argument_token= self.push(factory)
        assert isinstance(self._meta_argument_token, UnifiedToken)
        return self._meta_argument_token
    
    def _new_unified_token(self, factory: ArgumentFactory) -> UnifiedToken:
        return UnifiedToken(factor=factory.argument_glossary)
    
    def push(self, factory: ArgumentFactory) -> UnifiedToken:
        return super().push(factory)
    
    def pop(self, token: UnifiedToken, *args, **kwargs) -> Any:
        result = super().pop(token, *args, **kwargs)
        self._meta_argument_token = None
        return result
    
    def search(self, token: UnifiedToken, *args, **kwargs) -> ArgumentFactory | None:
        return super().search(token, *args, **kwargs)
    
    def reset(self) -> bool:
        result = super().reset()
        self._meta_argument_token = None
        return result

        



from pathlib import Path
from typing import List
from nlpe import DatasetProxy, Data, TextData, DatasetSplitCategory, DatasetVersion, Text, Language
import json

def test_pseudo_dataset():
    pseudo_dataset = {DatasetSplitCategory.TRAIN: ["tr1", "tr2", "tr3"], DatasetSplitCategory.VALIDATION: ["v1, v2"], DatasetSplitCategory.TEST: ["te1"]}
    proxy = DatasetProxy(
        glossary="pseudo_dataset",  
        dataset_type=list,
        load_dataset_call=lambda proxy_self, split, version: pseudo_dataset[split], 
        dump_dataset_call=lambda proxy_self, split, version: True, 
        raw_dir=Path("tmp", "raw_dir"))
    test_data = Data(dataset_proxy=proxy)
    all_datasets = test_data.load_dataset()
    assert len(all_datasets) == 3
    for s, d in zip(DatasetSplitCategory.all, all_datasets):
        assert id(pseudo_dataset[s]) == id(d)
        
    for s in DatasetSplitCategory.all:
        assert id(pseudo_dataset[s]) == id(proxy.load_dataset(s))
        
def test_pseudo_data_io():
    pseudo_dataset = {DatasetSplitCategory.TRAIN: ["tr1", "tr2", "tr3"], DatasetSplitCategory.VALIDATION: ["v1, v2"], DatasetSplitCategory.TEST: ["te1"]}
    pseduo_raw_file = lambda raw_dir: Path(raw_dir, "pseduo.json")
    pseduo_split_file = lambda split, version: Path("tmp", "pseduo", version, f"{split}.json")
    def load_dataset_call(proxy_self: DatasetProxy, split: DatasetSplitCategory, version:DatasetVersion, *args, **kwargs):
        raw_dir = proxy_self.raw_dir
        if version == DatasetVersion.RAW:
            print(f"load {split} from raw file: {pseduo_raw_file(raw_dir)}")
            return json.loads(pseduo_raw_file(raw_dir).read_text())
        elif version == DatasetVersion.FEATURE:
            if not pseduo_split_file(split, version).exists():
                print(f"load {split} from raw file: {pseduo_raw_file(raw_dir)}")
                return json.loads(pseduo_raw_file(raw_dir).read_text())[split]
            else:
                print(f"load {split} from pseduo file: {pseduo_split_file(split, version)}")
                return json.loads(pseduo_split_file(split, version).read_text())
        else:
            raise ValueError
        
    def dump_dataset_call(proxy_self: DatasetProxy, split: DatasetSplitCategory, version:DatasetVersion, *args, **kwargs):
        raw_dir = proxy_self.raw_dir
        if version == DatasetVersion.RAW:
            print(f"dump {split} to raw file: {pseduo_raw_file(raw_dir)}")
            pseduo_raw_file(raw_dir).write_text(json.dumps(pseudo_dataset, indent=4))
        else:
            print(f"dump {split} to pseduo file: {pseduo_split_file(split, version)}")
            pseduo_split_file(split, version).parent.mkdir(parents=True, exist_ok=True)
            pseduo_split_file(split, version).write_text(json.dumps(pseudo_dataset[split]))
        return True
    
    proxy = DatasetProxy(
        glossary="pseudo_dataset_io",  
        dataset_type=list,  
        download_raw_call= lambda raw_dir: pseduo_raw_file(raw_dir).write_text(json.dumps(pseudo_dataset, indent=4)), 
        load_dataset_call=load_dataset_call, 
        dump_dataset_call=dump_dataset_call, 
        raw_dir=Path("tmp", "raw_dir"))
    test_data = Data(dataset_proxy=proxy)
    all_datasets = test_data.load_dataset()
    assert len(all_datasets) == 3
    for s, d in zip(DatasetSplitCategory.all, all_datasets):
        assert pseudo_dataset[s] == d
        
    for s in DatasetSplitCategory.all:
        pseudo_dataset[s] == test_data[s]
    
    for s in DatasetSplitCategory.all:
        pseudo_dataset[s] == test_data[s]
    
    test_data.dump_dataset(version=DatasetVersion.VANILLA)
    test_data.dump_dataset(version=DatasetVersion.FEATURE)
    test_data.dump_dataset(version=DatasetVersion.RAW)    


def test_pseudo_text_data():
    pseudo_dataset = {DatasetSplitCategory.TRAIN: ["Hello World.", "How are You."], DatasetSplitCategory.VALIDATION: ["I am validataion."], DatasetSplitCategory.TEST: ["Nice to meet you!"]}
    proxy = DatasetProxy(
        glossary="pseudo_text_dataset",  
        dataset_type=list,
        load_dataset_call=lambda proxy_self, split, version: pseudo_dataset[split], 
        dump_dataset_call=lambda proxy_self, split, version: True, 
        raw_dir=Path("tmp", "raw_dir"))
    
    test_data = TextData(dataset_proxy=proxy, map_dataset_to_text_list=lambda dataset: [Text(s) for s in dataset])
    text_set = set(test_data.all_texts)
    assert len(text_set) == 4
    for s in DatasetSplitCategory.all:
        for string in pseudo_dataset[s]:
            assert string in text_set
    
    test_data.statistic_all_texts(tokenizor=str.split)
    assert test_data.max_length == 4
    assert test_data.min_length == 2
    


import json
import os.path
import sys
import random
from typing import Any, Callable, Dict, List, Optional, Iterable, Union
import numpy as np


def set_seed(seed=1):
    random.seed(seed)
    np.random.seed(seed)

def class_name_chain(cls: type):
    supers = cls.__mro__
    names = [c.__name__ for c in supers]
    return '-->'.join(names)


def jsonable(obj: Any):
    if isinstance(obj, dict):
        result = dict()
        for k, v in obj.items():
            if type(k) not in (str, int, float, bool, None):
                k=str(k)
            result[k] = jsonable(v)
        return result
    else:
        result = obj
        try:
            json.dumps(obj)
        except:
            result = str(obj)
        return result


def map_item_wise(map_fun, container, item_identify):
    def _map(item):
        if item_identify(item):
            return map_fun(item)

        if isinstance(item, Iterable):
            result = []
            for i in item:
                result.append(_map(i))
            result = type(item)(result)
            return result
        else:
            raise ValueError

    return _map(container)


def search_arg(name) -> bool:
    name = f'--{name}'
    args = sys.argv
    if name in args:
        return True
    return False


def set_up_args(args_file: str) -> None:
    assert os.path.isfile(args_file)
    with open(args_file) as af:
        args_dict: dict = json.load(af)
    flags = []
    for f in set(args_dict.get('flags', [])):
        if not search_arg(f):
            flags.append(f)
    for name, value in args_dict.items():
        if not search_arg(name):
            sys.argv.extend([f'--{name}', value])
    sys.argv.extend(map(lambda x: f'--{x}', flags))
    pass


def normalize_str_arg(arg: str):
    assert isinstance(arg, str)
    return arg.strip()


import json
from typing import Dict, List
from nlpe import ArgumentPool, DatasetProxy, TextData, Text, DatasetSplitCategory, Approach, DatasetVersion, EvaluatorProxy, ArgumentFactory
from nlpe.utils import global_logger, jsonable
from pathlib import Path
from transformers import BertForSequenceClassification, BertTokenizerFast
from transformers import Trainer, DataCollatorWithPadding, EvalPrediction, HfArgumentParser, TrainingArguments
from datasets import Dataset
from dataclasses import asdict
import torch
from functools import wraps
import random
from nlpe.utils.glossary import Glossary
import datasets
import evaluate
import numpy as np


class SigmoidClassifier(torch.nn.Linear):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.sigmoid(super().forward(input))
    
class SementicSimilarity(Approach):
    def _process(self, data: TextData, *args, **kwargs):
        logger.info("Load tokenizer of BERT")
        tokenizer = BertTokenizerFast.from_pretrained("google-bert/bert-base-uncased", cache_dir=Path(meta_args["cache_dir"], "model"))
        data.statistic_all_texts(tokenizor=lambda text: tokenizer(text)["input_ids"])
        logger.info(f"Max lengtg of {data.dataset_name} is: {data.max_length}")
        logger.info(f"Min lengtg of {data.dataset_name} is: {data.min_length}")
        logger.info("Load pretraind BERT")
        model = BertForSequenceClassification.from_pretrained("google-bert/bert-base-uncased", cache_dir=Path(meta_args["cache_dir"], "model"), num_labels=1, problem_type = "regression")
        config = model.config
        model.classifier.forward = SigmoidClassifier(config.hidden_size, config.num_labels)
        assert config.num_labels == 1
        def tokenize_input(samples:Dict):
            inputs = tokenizer(samples["sentence1"], samples["sentence2"])
            if "label" in samples:
                inputs["label"] = samples["label"]
            return inputs
        
        logger.info("Initialize trainer")
        trainer = Trainer(
            model=model,
            args=ArgumentPool()["trainer_argument"],
            data_collator=DataCollatorWithPadding(tokenizer),
            train_dataset=stsb_data.load_dataset(DatasetSplitCategory.TRAIN).map(tokenize_input),
            eval_dataset=stsb_data.load_dataset(DatasetSplitCategory.VALIDATION).map(tokenize_input),
            compute_metrics=data.evaluator_proxy.compute
        )
        logger.info("Training")
        trainer.train()
        logger.info("Evaluation")
        logger.info(trainer.evaluate())
        logger.info("Testing")
        logger.info(trainer.evaluate(stsb_data.load_dataset(DatasetSplitCategory.TEST).map(tokenize_input), metric_key_prefix="test"))
    
    
def load_dataset_call(proxy: DatasetProxy, split: DatasetSplitCategory, *args, **kwargs):
    logger.info(f"Load STSB dataset, split: {split}")
    return datasets.load_dataset("nyu-mll/glue", name="stsb", split=split, cache_dir=Path(proxy.raw_dir))


def dump_dataset_call(proxy: DatasetProxy, split: DatasetSplitCategory, *args, **kwargs):
    dataset_dir = Path("dataset")
    dataset_dir.mkdir(exist_ok=True)
    dataset: datasets.Dataset = stsb_data.dataset[split]
    version: DatasetVersion = kwargs["version"]
    if version is DatasetVersion.RAW:
        pass
    else:
        dataset.to_json(Path(datasets),to_json_kwargs=dict(intent=4))


def evaluate(*args, **kwargs):
    import scipy.stats
    from sklearn.metrics import mean_absolute_error, mean_squared_error
    # breakpoint()
    if len(args) == 0:
        predicts = kwargs["predicts"]
        labels = kwargs["labels"]
    elif len(args) == 1:
        eval_pre: EvalPrediction = args[0]
        assert isinstance(eval_pre, EvalPrediction)
        predicts = eval_pre.predictions
        labels = eval_pre.label_ids
    elif len(args) == 2:
        predicts, labels = args
    else:
        raise ValueError(f"{args} is not required")
    
    return {
        # "pearsonr_cor": scipy.stats.pearsonr(predicts, labels).pvalue,
        "pvalue": scipy.stats.spearmanr(predicts, labels).pvalue,
        "MAE": mean_absolute_error(predicts, labels),
        "MSE": mean_squared_error(predicts, labels)
    }
    
    
logger = global_logger()

# register the tranier arg to ArgumentPool
ArgumentPool().push(ArgumentFactory(
    argument_glossary="trainer_argument",
    argument_type= TrainingArguments,
    process=lambda : HfArgumentParser(TrainingArguments).parse_args_into_dataclasses(return_remaining_strings=True)[0], 
    to_dict=asdict))

logger.info("All arguments:")
logger.info(json.dumps(jsonable(ArgumentPool().all_args), indent=4))

# Synchronize random seed
seed = ArgumentPool()["trainer_argument"].seed
random.seed(seed)
np.random.seed(seed)
torch.manual_seed(seed)


# Prepare dataset
meta_args = ArgumentPool().meta_argument
dataset_proxy = DatasetProxy(
                             dataset_type=datasets.Dataset, 
                             load_dataset_call=load_dataset_call, 
                             dump_dataset_call=dump_dataset_call,
                             raw_dir=Path(meta_args["cache_dir"], "dataset"))
evaluator_Proxy = EvaluatorProxy(
    compute_call=evaluate
)
stsb_data = TextData(
    dataset_proxy=dataset_proxy, 
    map_dataset_to_text_list=lambda dataset: [Text(i) for i in dataset["sentence1"] + dataset["sentence2"]],
    evaluator_proxy= evaluator_Proxy)

# New approach
approach = SementicSimilarity()

# process data
approach.process(stsb_data)



import sys

if sys.version_info.major != 3 or sys.version_info.minor < 11:
    print(f"The current python version is: \n {sys.version} \nMinimum requirement: 3.11 !")
    raise RuntimeError("Python Version Error!")


import re
from typing import Optional





from .data import Data, DatasetProxy, TextData, Text, Language, DatasetSplitCategory, DatasetVersion
from .evaluator import EvaluatorProxy
from .approach import Approach
from .argument import ArgumentPool, ArgumentFactory, PATH_MODE
from .pool import Pool, POOL_DEFAULT_CAPACITY, UnifiedToken, TokenStatus






import re
from typing import Optional, Any, Union
from .utils import normalize_str_arg
from .design_patterns import singleton
from ..pool import Pool, UnifiedToken


class Glossary:
    def __init__(self, name: str, abbreviation: str='', description:str='', force=False) -> None:
        name = normalize_str_arg(name)
        abbreviation = normalize_str_arg(abbreviation)
        description = normalize_str_arg(description)
        if not re.search('.', name):
            raise ValueError(f'"name" can not be empty')
        if not re.search('.', abbreviation):
            abbreviation = name
        self._name = name
        self._abbreviation=abbreviation
        self._description=description
        GLOSSARY_POOL.push(self, force)

    uuid_type = str
    
    @property
    def name(self):
        return self._name
    
    @property
    def uuid(self):
        return self.name
    
    @property
    def description(self):
        self._description
    
    @description.setter
    def description(self, desc: Optional[str]):
        if (desc is not None) and (not re.search('.', desc)):
                raise ValueError(f'Argument "desc" ({desc}) is invalid')
        self._description=desc
    
    @property
    def abbreviation(self) -> str:
        return self._abbreviation

    
    def __hash__(self) -> int:
        return hash(self.uuid)
    

    def __eq__(self, value: object) -> bool:
        return self.uuid==value.uuid
    
    def __str__(self) -> str:
        return str(self.uuid)


@singleton
class GlossaryPool(Pool):
    def __init__(self):
        super().__init__(unit_type=Glossary)
        
    def _new_unified_token(self, unit: Union[Any, Glossary]) -> UnifiedToken:
        if isinstance(unit, Glossary.uuid_type):
            return UnifiedToken(factor=unit)
        else:
            assert isinstance(unit, Glossary)
            return UnifiedToken(factor=unit.uuid)
    
    def reset(self):
        raise RuntimeError("Glossary Pool Can not be reseted")
    
    def pop(self, token: UnifiedToken, *args, **kwargs) -> Any:
        raise RuntimeError("Can not delete unit from Glossary Pool")
    
    def push(self, glossary: Glossary, force=False) -> UnifiedToken:
        old = self.search(glossary)
        if old:
            if not force:
                raise ValueError(f"Glossary ({old}) has been exist! Overwrite the exsit glossary with current definiation by set 'force' to 'True'.")
            else:
                from .log import global_logger
                logger = global_logger()
                logger.warning(f"Glossary ({vars(old)}) has been exist! Overwrite with ({vars(glossary)})")
        token = super().push(glossary)
        return token
    
    def search(self, glossary: Union[Any, Glossary], *args, **kwargs) -> Any | None:
        token = self._new_unified_token(glossary)
        return super().search(token, *args, **kwargs)
    
GLOSSARY_POOL = GlossaryPool()


import uuid
from dataclasses import dataclass
from enum import Enum, StrEnum, auto
from typing import Any, Dict, List, Optional, Set, Union
from uuid import uuid4 as random_uuid, UUID


class PoolState(Enum):
    FREE = auto()
    BUSY = auto()
    EMPTY = auto()
    FULL = auto()


class TokenStatus(StrEnum):
    VALID = auto()
    INVALID = auto()


class UnifiedToken:
    def __init__(self, factor = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if factor is None:
            factor = random_uuid()
        self._factor = factor
        self._status = TokenStatus.VALID
    
    @property
    def status(self):
        return self._status
    
    @status.setter
    def status(self, value):
        assert isinstance(value, TokenStatus)
        self._status = value

    @property
    def factor(self):
        return self._factor

    def __eq__(self, value: object) -> bool:
        assert isinstance(value, UnifiedToken)
        return self.factor == value.factor
    
    def __hash__(self):
        return hash(self.factor)
    
    def __str__(self) -> str:
        return str(self.factor)
    
    def __repr__(self) -> str:
        return repr(self.factor)



POOL_DEFAULT_CAPACITY = 128


class Pool:
    def __init__(self, unit_type: type = object):
        self._state: PoolState = PoolState.FREE
        self._capacity: int = POOL_DEFAULT_CAPACITY
        assert isinstance(unit_type, type)
        self._unit_type = unit_type
        self._utoken2unit: Dict[UnifiedToken, self._unit_type] = dict()

    @property
    def size(self) -> int:
        return len(self._utoken2unit)
        
    @property
    def _all_units(self) -> List[Any]:
        return list(self._utoken2unit.values())
    
    def _new_unified_token(self, unit: Any) -> UnifiedToken:
        return UnifiedToken()

    def push(self, unit: Any) -> UnifiedToken:
        assert self.size < POOL_DEFAULT_CAPACITY
        assert isinstance(unit, self._unit_type)
        united_token = self._new_unified_token(unit)
        self._utoken2unit[united_token] = unit
        united_token.status = TokenStatus.VALID
        return united_token

    def pop(self, token: UnifiedToken, *args, **kwargs) -> Any:
        assert isinstance(token, UnifiedToken)
        assert token.status == TokenStatus.VALID
        result = self._utoken2unit.pop(token)
        token.status = TokenStatus.INVALID
        return result

    def search(self, token: UnifiedToken, *args, **kwargs) -> Optional[Any]:
        assert isinstance(token, UnifiedToken)      
        assert token.status == TokenStatus.VALID  
        result = self._utoken2unit.get(token, None)
        return result
    
    def reset(self) -> bool:
        for ut in self._utoken2unit:
            ut.status = TokenStatus.INVALID
        self._utoken2unit.clear()
        return True



from nlpe import Data, Approach


class PseduoApproach(Approach):
    def _process(self, data: Data, *args, **kwargs):
        pass
    
    
def test_pseduo_approach():
    approach = PseduoApproach(glossary="pseduo_approach")
    assert approach.processing_data == None
    assert approach.glossary.name == "pseduo_approach"


from copy import deepcopy
import sys
from argparse import ArgumentParser
from typing import Any, Callable
from nlpe import ArgumentFactory, ArgumentPool, Approach, EvaluatorProxy, Data, DatasetProxy, DatasetSplitCategory, DatasetVersion
from nlpe.utils import global_logger, Glossary
from nlpe.utils.test import extend_to_original_sys_argv

sys.argv.extend("-t simple_task -d simple_data -a simple_approach --debug --evaluator simple_evaluator".split())
print("Current sys.argv: ", sys.argv)

parser = ArgumentParser()
parser.add_argument("--evaluator", type=str)
arg_pool = ArgumentPool()
arg_pool.push(ArgumentFactory(process=lambda : vars(parser.parse_known_args[0]), argument_glossary=Glossary("my_argument"), argument_type=dict))

logger = global_logger()

class SimpleApproach(Approach):
    
    def _process(self, data: Data, evaluator_proxy: EvaluatorProxy, *args, **kwargs):
        logger.info(f"Processing dataset: {data.dataset_proxy.glossary}")
        logger.info(f"Dataset: {data.all_datasets}")
        
        data.load_dataset(DatasetSplitCategory.TRAIN)
        logger.info(f"Load Model")
        logger.info(f"Trainning")
        logger.info(f"Dump Model")
        
        data.load_dataset(DatasetSplitCategory.VALIDATION)
        logger.info(f"Evaluating")
        evaluation = evaluator_proxy.compute([1,2,3],[3,3,3])
        logger.info(f"Evaluation Result for Valid Set:")
        logger.info(evaluation)
        logger.info(f"Dump Evaluation Result for Valid Set")
        
        data.load_dataset(DatasetSplitCategory.TEST)
        logger.info(f'Testing')
        evaluation = evaluator_proxy.compute([1,2,4],[3,2,3])
        logger.info(f"Evaluation Result for Test Set:")
        logger.info(evaluation)
        logger.info(f"Dump Evaluation Result for Test Set")


def setup_function(function):
    extend_to_original_sys_argv("-t test_task -d test_data -a test_approach --debug")
    
def test_approach_pipeline():
    approach = SimpleApproach(glossary="simple_approach")
    dataset_proxy = DatasetProxy(glossary="simple_dataset_proxy", dataset_type=list, load_dataset_call=lambda *args, **kwargs: [1,2,3], dump_dataset_call=lambda *args, **kwargs: True)
    evaluator_proxy = EvaluatorProxy(glossary="simple_evaluator_proxy", compute_call=lambda pridiction, label: [*pridiction, *label])
    data = Data(dataset_proxy, evaluator_proxy)
    assert approach.processing_data is None
    approach.process(data, evaluator_proxy)
    assert approach.processing_data is None
